from ._fixtures import *

from s6r_odoo.odoo import OdooConnection
from s6r_odoo.model import OdooModel
from s6r_odoo.record import OdooRecord
from s6r_odoo.record_set import OdooRecordSet

def test_model(odoo):
    model_name = 'res.partner'
    model = odoo.model(model_name)

    assert model_name in odoo._models
    assert isinstance(model,OdooModel)

def test_get_object_reference(odoo):
    object_reference = odoo.get_object_reference('base.module_account')

    assert len(object_reference) == 2
    assert object_reference[0] == 'ir.module.module'
    assert isinstance(object_reference[1], int)

def test_get_ref(odoo):
    res_id = odoo.get_ref('base.module_account')

    assert isinstance(res_id,int)

def test__read_success(odoo_multi_versions):
    odoo = odoo_multi_versions
    ids = [1]
    fields = [
        'author',
        'description',
        'display_name',
        'name',
        'state',
        'website',
    ]
    res = odoo._read('ir.module.module',ids,fields)

    assert len(res)==1
    assert res[0]['author'] == 'Odoo S.A.'

def test_ref(odoo_multi_versions):
    #trying ref() on 'ir.module.module' model raises an XMPLRPC error on v17+ (because of None value)
    odoo = odoo_multi_versions
    country_fr = odoo.ref('base.fr')

    assert isinstance(country_fr, OdooRecord)
    assert country_fr.name == 'France'
    #TODO
    #assert country_fr._xmlid = 'base.fr'

def test_results_instance(odoo_multi_versions):
    odoo = odoo_multi_versions
    model_id = odoo.model('ir.model').read([1], fields=['name'])
    assert isinstance(model_id, OdooRecordSet)
    model_ids = odoo.model('ir.model').read([1, 2, 3], fields=['name'])
    assert isinstance(model_ids, OdooRecordSet)
    model_ids = odoo.model('ir.model').search([('name', '=', 'test')], fields=['name'])
    assert isinstance(model_ids, OdooRecordSet)
    model_ids = odoo.model('ir.model').search([('name', '=', 'Base')], fields=['name'])
    assert isinstance(model_ids, OdooRecordSet)
    partner_id = odoo.model('res.partner').create([{'name': 'Test', 'email': 'test@test.fr'}])
    assert isinstance(partner_id, OdooRecordSet)
    partner_id = odoo.model('res.partner').create({'name': 'Test', 'email': 'test@test.fr'})
    assert isinstance(partner_id, OdooRecordSet)

    # Search with limit=1 returns a record
    model_id = odoo.model('ir.model').search([], fields=['name'], limit=1)
    assert isinstance(model_id, OdooRecord)
    model_id = odoo.model('ir.model').search([('name', '=', 'test')], fields=['name'], limit=1)
    assert isinstance(model_id, OdooRecordSet) and len(model_id) == 0



# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from .record import OdooRecord

class OdooRecordSet(list):
    _odoo = None
    _model = None

    def __init__(self, seq=(), model=None):
        super().__init__(seq)
        if model:
            self._model = model
            self._odoo = self._model._odoo

    @property
    def ids(self):
        return self.get_ids()

    def to_dict(self):
        """Convert the recordset to a list of dictionaries for JSON serialization"""
        return [record.to_dict() if hasattr(record, 'to_dict') else dict(record) for record in self]
        
    @property
    def _read_fields(self):
        read_fields = set()
        for r in self:
            read_fields.update(r._read_fields)
        return list(read_fields)

    def save(self, batch_size=100, skip_line=0, ignore_fields=None, context=None):
        ignore_fields = ignore_fields or []

        values_list = [r.get_update_values() for r in self]
        res = self._model.load_batch(values_list, batch_size=batch_size,skip_line=skip_line,
                                     ignore_fields=ignore_fields, context=context)
        res_ids = res.get('ids', False)
        if not res_ids:
            return False
        for i, r in enumerate(self):
            #remove from updated values the fields that have been updated
            r._updated_values = {k:v for k,v in r._updated_values.items() if k in ignore_fields}

            #ideally, '/id' should never end up in _values, WIP
            if '/id' in r._values:
                r._values.pop('/id')
            # initialized_fields should only contain existing field names (e.g. no field names like 'user_id/id')
            r._initialized_fields.extend([k.replace('/id','') for k in r._values.keys() if k not in ignore_fields])
            #remove duplicates
            r._initialized_fields = list(set(r._initialized_fields))

            if not r.id:
                r.id = res_ids[i]
        return True

    def get_ids(self):
        return [r.id for r in self]

    def unlink(self):
        ids = self.ids
        self._model.unlink(ids)
        for key in ids:
            self._model._cache.pop(key, None)
        self.clear()

    def filtered(self, func=None, **kwargs):
        if func:
            res = [r for r in self if func(r)]
        else:
            res = [r for r in self if all(getattr(r, k) == v for k, v in kwargs.items())]
        return OdooRecordSet(res, model=self._model)

    def mapped(self, path):
        res = []
        parents = self
        for field in path.split('.'):
            res = set([getattr(r, field) for r in parents if r])
            parents = res
        res = list(res)
        if res and isinstance(res[0], OdooRecord):
            return OdooRecordSet(res, model=res[0]._model)
        return res



try:
    from ._fixtures import *
except ImportError:
    pass

def test_csv_file_import(odoo):
    model_name = 'res.partner'
    result = odoo.import_csv('res.partner.csv', model_name, limit=100, skip_line=500, batch_size=50)
    _test_import_result(odoo, model_name, result)

def _test_import_result(odoo, model_name, result):
    assert result and result.get('ids')
    assert len(result.get('ids')) == 100
    partner_id = odoo.model(model_name).read(result.get('ids')[0], ['name'])
    assert partner_id.name == 'Partner 501'
    partner_id = odoo.model(model_name).read(result.get('ids')[-1], ['name'])
    assert partner_id.name == 'Partner 600'

def test_xls_file_import(odoo):
    model_name = 'res.partner'
    result = odoo.import_xls('res.partner.xls', model_name, limit=100, skip_line=500, batch_size=50)
    _test_import_result(odoo, model_name, result)

def test_xlsx_file_import(odoo):
    model_name = 'res.partner'
    result = odoo.import_xlsx('res.partner.xlsx', model_name, limit=100, skip_line=500, batch_size=50)
    _test_import_result(odoo, model_name, result)



from ._fixtures import *

from s6r_odoo.odoo import OdooConnection
from s6r_odoo.model import OdooModel
from s6r_odoo.record import OdooRecord
from s6r_odoo.record_set import OdooRecordSet

@pytest.mark.skip(reason='WIP')
def test_read_without_fields_without_cache(odoo):

    partner_values = {
        # 'id':1,
        'name': 'Jean-Claude Test',
        'email': 'jc@test.com',
    }
    partner_id = odoo.values_to_record('res.partner', partner_values)

    #doesn't return anything
    p = partner_id.read(no_cache=True)

    pass

    assert False



# Copyright (C) 2025 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import csv
import os
from datetime import datetime
import sys
import xlrd
import openpyxl

def get_file_full_path(path):
    if not path:
        return ''
    param_path = path
    if not os.path.isfile(path):
        path = os.path.join(os.path.dirname(sys.argv[0]), param_path)
    if not os.path.isfile(path):
        raise FileNotFoundError('%s not found!' % param_path)
    return path

class FileImport:
    name_create_enabled_fields = ''
    context = False
    limit = 0
    skip_line = 0
    batch_size = 100
    thread = 0
    ignore_errors = []

    def __init__(self, odoo, **kwargs):
        self.odoo = odoo
        self.logger = odoo.logger
        self._context = odoo._context.copy()
        self.set_params(kwargs)

    @staticmethod
    def parse_csv_file(file_path, delimiter=","):
        vals = []
        with open(os.path.dirname(__file__) + '/' + file_path, 'r') as csvfile:
            reader = csv.reader(csvfile, skipinitialspace=True, delimiter=delimiter)
            next(reader)
            for line in reader:
                vals.append(line)
        return vals

    @staticmethod
    def clean_field(field):
        if "/" in field:
            field = field.split("/")[0]
        if "." in field:
            field = field.split(".")[0]
        return field

    @staticmethod
    def field_check_integer(field, data):
        if field['name'] == 'id':
            return data
        if not data:
            return data
        data = int(data)
        return data

    @staticmethod
    def field_check_float(field, data):
        if not data:
            return data
        if isinstance(data, str) and "," in data:
            data = data.replace(',', '.')
        data = float(data)
        return data

    @staticmethod
    def field_check_monetary(field, data):
        return FileImport.field_check_float(field, data)

    @staticmethod
    def field_check_date(field, data):
        if not data:
            return data
        format_ok = False
        try:
            datetime.strptime(data, '%Y-%m-%d')
            format_ok = True
        except:
            format_ok = False
        if not format_ok:
            try:
                data = datetime.strptime(data, '%d/%m/%Y').strftime('%Y-%m-%d')
                format_ok = True
            except Exception as e:
                format_ok = False
        return data

    @staticmethod
    def field_check_datetime(field, data):
        if not data:
            return data
        format_ok = False
        try:
            datetime.strptime(data, '%Y-%m-%d %H:%M:%S')
            format_ok = True
        except Exception as e:
            format_ok = False
        if not format_ok:
            try:
                data = datetime.strptime(data, '%d/%m/%Y %H:%M:%S').strftime('%Y-%m-%d %H:%M:%S')
                format_ok = True
            except Exception as e:
                format_ok = False
        if not format_ok:
            data = FileImport.field_check_date(field, data)
            data += " 00:00:00"

        return data

    def parse_csv_file_dictreader(self, file_path, fields, delimiter=","):
        vals = []
        file_path = get_file_full_path(file_path)
        with open(file_path, 'r') as csvfile:
            header = csvfile.readline()
            csvfile.seek(0, 0)
            if delimiter not in header:
                if ';' in header:
                    delimiter = ';'
            reader = csv.DictReader(csvfile, skipinitialspace=True, delimiter=delimiter, quotechar='"')
            cc = 0
            line_count = 0
            for line in reader:
                if self.skip_line and cc < self.skip_line:
                    cc += 1
                    continue
                if self.limit and line_count == self.limit:
                    break
                cc += 1
                line_count += 1
                for field in line:
                    clean_field = FileImport.clean_field(field)
                    method = "field_check_%s" % (fields.get(clean_field,{}).get('type'))
                    if hasattr(FileImport, method):
                        method_to_call = getattr(FileImport, method)
                        line[field] = method_to_call(fields[clean_field], line[field])
                vals.append(line)
        return vals

    def set_params(self, params):
        self.name_create_enabled_fields = params.get('name_create_enabled_fields', '')
        self.context = params.get('context', False)
        self.limit = params.get('limit', 0)
        self.skip_line = params.get('skip_line', 0)
        self.batch_size = params.get('batch_size', 100)
        self.thread = params.get('thread', 0)
        self.ignore_errors = params.get('ignore_errors', [])

        if self.name_create_enabled_fields:
            create_enabled_fields = dict([x, 1] for x in set(self.name_create_enabled_fields.split(",")))
            self._context.update({'name_create_enabled_fields': create_enabled_fields})
        if self.context:
            self._context.update(self.context)

    def import_csv(self, file_path, model, **kwargs):
        self.set_params(kwargs)
        fields = self.odoo.get_fields(model)
        raw_datas = self.parse_csv_file_dictreader(file_path, fields)
        if raw_datas:
            return self.odoo.load_batch(model, raw_datas, batch_size=self.batch_size)

    def _xls_rows_to_values(self, rows, fields):
        values = []
        headers = rows.pop(0)
        for row in rows:
            value = {}
            for header in [h for h in headers if h]:
                value[header] = row[headers.index(header)]
            values.append(self.check_values(value, fields))
        if self.skip_line:
            values = values[self.skip_line:]
        if self.limit:
            values = values[:self.limit]
        return values

    def check_values(self, values, fields):
        for field in values:
            clean_field = FileImport.clean_field(field)
            method = "field_check_%s" % (fields.get(clean_field, {}).get('type'))
            if hasattr(FileImport, method):
                method_to_call = getattr(FileImport, method)
                values[field] = method_to_call(fields[clean_field], values[field])
        return values

    def read_xls(self, file_path):
        file_path = get_file_full_path(file_path)
        book = xlrd.open_workbook(file_path)
        sheets = book.sheet_names()
        sheet = sheets[0]
        return self._read_xls_book(book, sheet)

    def read_xlsx(self, file_path):
        file_path = get_file_full_path(file_path)
        book = openpyxl.load_workbook(file_path)
        sheets = book.sheetnames
        sheet = sheets[0]
        return self._read_xlsx_book(book, sheet)

    def _read_xlsx_book(self, book, sheet_name):
        sheet = book.get_sheet_by_name(sheet_name)
        rows = []
        for row in sheet.rows:
            values = []
            for colx, cell in enumerate(row, 1):
                if cell.data_type == 'n' and cell.value is not None:
                    is_float = cell.value % 1 != 0.0
                    values.append(str(cell.value) if is_float else str(int(cell.value)))
                elif cell.data_type == 'd':
                    dt = cell.value
                    value = dt.strftime("%Y-%m-%d %H:%M:%S") if isinstance(dt, datetime) else dt.strftime("%Y-%m-%d")
                    values.append(value)
                elif cell.data_type == 'b':
                    values.append(u'True' if cell.value else u'False')
                elif cell.value is not None:
                    values.append(cell.value)
            if any(x for x in values if x.strip()):
                rows.append(values)
        return rows

    def _read_xls_book(self, book, sheet_name):
        sheet = book.sheet_by_name(sheet_name)
        rows = []
        for row in map(sheet.row, range(sheet.nrows)):
            values = []
            for cell in row:
                if cell.ctype is xlrd.XL_CELL_NUMBER:
                    is_float = cell.value % 1 != 0.0
                    value = str(cell.value) if is_float else str(int(cell.value))
                    values.append(value)
                elif cell.ctype is xlrd.XL_CELL_DATE:
                    is_datetime = cell.value % 1 != 0.0
                    dt = datetime(*xlrd.xldate.xldate_as_tuple(cell.value, book.datemode))
                    value = dt.strftime("%Y-%m-%d %H:%M:%S") if is_datetime else dt.strftime("%Y-%m-%d")
                    values.append(value)
                elif cell.ctype is xlrd.XL_CELL_BOOLEAN:
                    values.append(u'True' if cell.value else u'False')
                else:
                    values.append(cell.value)
            if any(x for x in values if x.strip()):
                rows.append(values)

        return rows

    def import_xls(self, file_path, model, **kwargs):
        self.set_params(kwargs)
        fields = self.odoo.get_fields(model)
        rows = self.read_xls(file_path)
        raw_datas = self._xls_rows_to_values(rows, fields)
        if raw_datas:
            return self.odoo.load_batch(model, raw_datas, batch_size=self.batch_size)

    def import_xlsx(self, file_path, model, **kwargs):
        self.set_params(kwargs)
        fields = self.odoo.get_fields(model)
        rows = self.read_xlsx(file_path)
        raw_datas = self._xls_rows_to_values(rows, fields)
        if raw_datas:
            return self.odoo.load_batch(model, raw_datas, batch_size=self.batch_size)



# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import base64
from datetime import datetime
import logging
import time
import ssl
import os
import sys
import xmlrpc.client
from pprint import pformat

import requests
from bs4 import BeautifulSoup

from .file_import import FileImport
from .model import OdooModel
from .record import OdooRecord
from .record_set import OdooRecordSet

METHODE_MAPPING = {
    15: [('get_object_reference', 'check_object_reference')]
}


class OdooConnection:
    query_count = 0
    method_count = {}

    def __init__(self, url, dbname, user, password, version=15.0, http_user=None, http_password=None, createdb=False,
                 debug_xmlrpc=False, legacy=False, logger=None, **kwargs):
        self.logger = logger or logging.getLogger("Odoo Connection".ljust(15))
        if debug_xmlrpc:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        self._url = url
        self._dbname = dbname
        self._user = user
        self._password = password
        self._http_user = http_user
        self._http_password = http_password
        self._version = version
        self._legacy = legacy
        self._debug_xmlrpc = debug_xmlrpc
        self._context = {'lang': kwargs.get('lang') or 'fr_FR', 'noupdate': True}
        # noinspection PyProtectedMember,PyUnresolvedReferences
        self._insecure_context = ssl._create_unverified_context()
        self._compute_url()
        if createdb:
            self._create_db()
        self._prepare_connection()
        self._models = {}
        self._file_import = FileImport(self)

    def __str__(self):
        return 'OdooConnection(%s)' % self._dbname

    def __repr__(self):
        return str(self)

    @property
    def context(self):
        return self._context

    def model(self, model_name):
        if model_name in self._models:
            return self._models[model_name]
        res = OdooModel(self, model_name)
        self._models[model_name] = res
        return res

    def _compute_url(self):
        if self._http_user or self._http_password:
            self._url = self._url.replace('https://', 'https://%s:%s@' % (self._http_user, self._http_password))

    def _get_xmlrpc_method(self, method):
        new_method = method
        for v in METHODE_MAPPING:
            if self._version >= v:
                for i in METHODE_MAPPING[v]:
                    if i[0] == method:
                        new_method = i[1]
        return new_method

    def init_logger(self, name='s6r-odoo', level='INFO'):
        root_logger = logging.getLogger()
        root_logger.setLevel(level)
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter(f'%(asctime)s - {name} - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)
        self.logger = root_logger

    def _create_db(self):
        post = {
            'master_pwd': "admin123",
            'name': self._dbname,
            'login': self._user,
            'password': self._password,
            'phone': '',
            'lang': 'fr_FR',
            'country_code': 'fr',
        }
        session = requests.Session()
        session.verify = False
        r = session.post(url=self._url + "/web/database/create", params=post)
        soup = BeautifulSoup(r.text, 'html.parser')
        alert = soup.find('div', attrs={"class": u"alert alert-danger"})
        if alert:
            self.logger.debug(self._url + "/web/database/create")
            self.logger.debug(post)
            self.logger.debug(alert.get_text())
            if "already exists" not in alert.text:
                raise Exception(alert.text)

    def _prepare_connection(self):
        self.logger.info("Prepare connection %s %s %s" % (self._url, self._dbname, self._user))
        self.common = xmlrpc.client.ServerProxy('{}/xmlrpc/2/common'.format(self._url), allow_none=True,
                                                context=self._insecure_context)
        self.object = xmlrpc.client.ServerProxy('{}/xmlrpc/2/object'.format(self._url), allow_none=True,
                                                context=self._insecure_context)
        self.logger.info("==============")
        try:
            self.uid = self.common.authenticate(self._dbname, self._user, self._password, {})
        except xmlrpc.client.Fault as err:
            if 'FATAL:' in err.faultString:
                msg = err.faultString[err.faultString.find('FATAL:') + 6:].strip()
                self.logger.error(msg)
                raise ConnectionError(msg)
            raise err
        except xmlrpc.client.ProtocolError as err:
            msg = f'{err.url} {err.errmsg} ({err.errcode})'
            self.logger.error(msg)
            raise ConnectionError(msg)
        except Exception as err:
            self.logger.error(err)
            raise

        self.logger.debug('Connection uid : %s' % self.uid)
        if not self.uid:
            msg = f'Connection Error to {self._url} {self._dbname}. Check "{self._user}" username and password.'
            self.logger.error(msg)
            raise ConnectionError(msg)

    def reset_count(self):
        self.query_count = 0
        self.method_count = {}

    def query_counter_update(self, model, method):
        self.query_count += 1
        if method not in self.method_count:
            self.method_count[method] = {model: 1}
        else:
            if model not in self.method_count[method]:
                self.method_count[method][model] = 1
            else:
                self.method_count[method][model] += 1

    def execute_odoo(self, *args, no_raise=False, no_log=False, retry=0):
        model = args[0]
        method = args[1]
        self.query_counter_update(model, method)

        self.logger.debug("*" * 50)
        self.logger.debug("Execute odoo :")
        self.logger.debug("\t Model : %s" % model)
        self.logger.debug("\t Method : %s" % method)
        self.logger.debug("\t " + "%s " * (len(args) - 2) % args[2:])
        self.logger.debug("*" * 50)
        try:
            if retry:
                self._prepare_connection()
            res = self.object.execute_kw(self._dbname, self.uid, self._password, *args)
            return res
        except (ConnectionResetError, ConnectionError, xmlrpc.client.ProtocolError) as e:
            if not retry:
                self.logger.info("Retry #1 to connect to Odoo...")
                time.sleep(5)
                return self.execute_odoo(*args, no_raise=no_raise, no_log=no_log, retry=1)
            elif retry == 1:
                self.logger.info("Retry #2 to connect to Odoo...")
                time.sleep(10)
                return self.execute_odoo(*args, no_raise=no_raise, no_log=no_log, retry=2)
            else:
                self.logger.error("Max connection retry reached.", exc_info=True)
                if not no_raise:
                    raise

        except Exception as e:
            if no_raise:
                pass

            if not no_log:
                self.logger.error(pformat(args))
                if hasattr(e, 'faultString'):
                    self.logger.error(e.faultString)
                    if 'InterfaceError: cursor already closed' in e.faultString:
                        if retry > 5:
                            raise e
                        self.logger.info("Retry to connect to Odoo...")
                        time.sleep(5)
                        return self.execute_odoo(*args, no_raise=no_raise, no_log=no_log, retry=retry+1)
                else:
                    self.logger.error(e)
            if not no_raise:
                raise e

    def values_to_record(self, model_name, values, update_cache=True, resolve_xmlids=True):
        if isinstance(values, int):
            values = {'id': values}
        record = OdooRecord(self, self.model(model_name), values, resolve_xmlids=resolve_xmlids)
        if update_cache:
            self._models[model_name]._update_cache(record.id, values)
        return record

    def values_list_to_records(self, model_name, val_list, update_cache=True, resolve_xmlids=True):
        if val_list is None:
            val_list = []
        if self._legacy:
            return val_list
        records = [self.values_to_record(model_name, values, update_cache, resolve_xmlids) for values in val_list]
        return OdooRecordSet(records, model=self.model(model_name))

    def get_ref(self, external_id):
        object_ref = self.get_object_reference(external_id)
        res = object_ref[1] if object_ref else False
        self.logger.debug('Get ref %s > %s' % (external_id, res))
        return res

    def get_local_file(self, path, encode=False):
        if encode:
            with open(path, "rb") as f:
                res = f.read()
                res = base64.b64encode(res).decode("utf-8", "ignore")
        else:
            with open(path, "r") as f:
                res = f.read()
        return res

    def get_country(self, code):
        return self.execute_odoo('res.country', 'search', [[('code', '=', code)], 0, 1, "id", False],
                                 {'context': self._context})[0]

    def get_menu(self, website_id, url):
        return self.execute_odoo('website.menu', 'search',
                                 [[('website_id', '=', website_id), ('url', '=', url)], 0, 1, "id", False],
                                 {'context': self._context})[0]

    def get_search_id(self, model, domain):
        return self.execute_odoo(model, 'search', [domain, 0, 1, "id", False], {'context': self._context})[0]

    def search_ir_model_data(self, domain):
        return self.model('ir.model.data').search(domain,
                                                  fields=['module', 'name', 'model', 'res_id'],
                                                  order='id')

    def get_object_reference_legacy(self, xml_id, no_raise=False):
        object_reference = self._get_xmlrpc_method('get_object_reference')
        return self.execute_odoo('ir.model.data', object_reference, xml_id.split('.'), no_raise=no_raise)


    def get_object_reference(self, xml_id, no_raise=False, cache_only=False):
        if self._legacy:
            return self.get_object_reference_legacy(xml_id, no_raise=no_raise)

        ref = self._get_object_reference_cache(xml_id)
        if ref:
            return ref
        if cache_only:
            return
        module, name = xml_id.split('.')
        domain = [('module', '=', module), ('name', '=', name)]
        res = self.search_ir_model_data(domain)
        if res:
            ir_model_data_id = res[0]
            return [ir_model_data_id.model, ir_model_data_id.res_id]

    def get_id_from_xml_id(self, xml_id, no_raise=False, cache_only=False):
        if '.' not in xml_id:
            xml_id = "external_config." + xml_id
        try:
            res = self.get_object_reference(xml_id, no_raise=no_raise, cache_only=cache_only)
            return res[1] if res else False
        except xmlrpc.client.Fault as fault:
            if no_raise:
                pass
            raise ValueError(fault.faultString.strip().split('\n')[-1])
        except Exception as err:
            if no_raise:
                pass
            else:
                raise err

    def ref(self, ixml_id, no_raise=False, fields=None, no_cache=False):
        object_reference = self.get_object_reference(ixml_id, no_raise=no_raise)
        if len(object_reference) == 2:
            model, res_id = object_reference
            res = self.model(model).read(res_id, fields=fields, no_cache=no_cache)
            res._xmlid = ixml_id
            return res

    def get_xml_id_from_id(self, model, res_id, cache_only=False):
        cache_xmlid = self._get_xmlid_cache(model, res_id)
        if cache_xmlid:
            return cache_xmlid
        if cache_only:
            return
        try:
            domain = [('model', '=', model), ('res_id', '=', res_id)]
            res = self.search_ir_model_data(domain)
            if res:
                datas = res[0]
                return "%s.%s" % (datas['module'], datas['name'])
            else:
                raise ValueError('xml_id not found.')
        except Exception as err:
            raise err

    def write(self, model, ids, values, context=None):
        return self.execute_odoo(model, 'write', [ids, values],
                                 {'context': context or self._context})

    def set_active(self, is_active, model, domain, search_value_xml_id):
        if search_value_xml_id:
            object_id = self.get_id_from_xml_id(search_value_xml_id)
            domain = [(domain[0][0], domain[0][1], object_id)]
        object_ids = self.search_ids(model, domain, context=self._context)
        self.write(model, object_ids, {'active': is_active})

    def read_search(self, model, domain, context=None):
        res = self.execute_odoo(model, 'search_read', [domain],
                                {'context': context or self._context})
        return res

    def search_count(self, model, domain, context=None):
        res = self.execute_odoo(model, 'search_count', [domain],
                                {'context': context or self._context})
        return res

    def _read(self, model, ids, fields, context=None):
        param_ids = [ids] if isinstance(ids, int) else ids
        return self.execute_odoo(model, 'read', [param_ids, fields],
                                 {'context': context or self._context})

    def read(self, model, ids, fields, context=None):
        if ids and isinstance(ids, list) and isinstance(ids[0], str):
            ids = [self.get_ref(i) for i in ids]
        res =  self._read(model, ids, fields, context)
        return self.values_list_to_records(model, res)

    def read_group(self, model, domain, fields, groupby, offset=0, limit=None, orderby=False, lazy=True, context=None):
        res = self.execute_odoo(model, 'read_group', [domain, fields, groupby, offset, limit, orderby, lazy],
                                {'context': context or self._context})
        return res

    def search(self, model, domain=[], fields=[], order="", offset=0, limit=0, context=None, **kwargs):
        if 'exclude_fields' in kwargs:
            exclude_fields = kwargs['exclude_fields']
            fields = [f for f in fields if f not in exclude_fields]
        if not fields:
            model_fields = self.get_fields(model)
            if len(model_fields) > 20:
                self.logger.warning(
                    f"You are trying to search and read {len(model_fields)} fields for model {model}"
                    "\nThis might slow down your script, consider using fields parameter.")
        params = [domain, fields, offset, limit, order]
        res = self.execute_odoo(model, 'search_read', params, {'context': context or self._context})
        if res and limit==1 and not self._legacy and not 'legacy' in kwargs:
            return self.values_list_to_records(model, res)[0]
        return self.values_list_to_records(model, res)

    def search_ids(self, model, domain=[], order="", offset=0, limit=0, context=None):
        params = [domain, offset, limit, order]
        res = self.execute_odoo(model, 'search', params, {'context': context or self._context})
        return res

    def get_record(self, model, rec_id, context=None):
        params = [[('id', '=', rec_id)]]
        res = self.execute_odoo(model, 'search_read', params,
                                {'context': context or self._context})
        if res:
            return res[0]

    def default_get(self, model, field):
        res = self.execute_odoo(model, 'default_get', [field])
        return res

    def load(self, model, load_keys, load_data, context):
        res = self.execute_odoo(model, 'load', [load_keys, load_data], {'context': context or self._context})
        for message in res['messages']:
            self.logger.error("%s : %s" % (message.get('record', False), message['message']))
        return res

    def prepare_load_batch_datas(self, datas):
        batch_keys = list(datas[0].keys())
        for i, values in enumerate(datas):
            for key in batch_keys:
                if key not in values:
                    self.logger.warning(f"Value for key {key} not found in line {i}")
                    values[key] = ''
                if key.endswith('.id') and isinstance(values.get(key), list):
                    values[key] = ','.join([str(v) for v in values.get(key)]) if values.get(key) else ''
                if key.endswith('.id') and isinstance(values.get(key), int):
                    values[key] = str(values.get(key))
                if isinstance(values[key], bool):
                    values[key] = '1' if values[key] else '0'
        return datas

    def load_batch(self, model, datas, batch_size=100, skip_line=0, context=None, **kwargs):
        ignore_fields = kwargs.get('ignore_fields', [])
        datas = datas[skip_line:]
        context = self.context | context if context else self.context
        if not datas:
            return
        cc_max = len(datas)
        start = datetime.now()

        load_keys = list(datas[0].keys())
        for field in ignore_fields:
            try:
                load_keys.remove(field)
            except ValueError:
                self.logger.warning(f"\"{field}\" field name not found in data keys")
        load_datas = [[]]
        datas = self.prepare_load_batch_datas(datas)
        for cc, data in enumerate(datas):
            if len(load_datas[-1]) >= batch_size:
                load_datas.append([])
            load_datas[-1].append([data[i] for i in load_keys])

        cc = 0
        res = {'ids': [], 'messages': []}
        for load_data in load_datas:
            start_batch = datetime.now()
            self.logger.info("\t\t* %s : %s-%s/%s" % (model, skip_line + cc, skip_line + cc + len(load_data), skip_line + cc_max))
            cc += len(load_data)
            load_res = self.load(model, load_keys, load_data, context=context)
            res['ids'].extend(load_res.get('ids', []) or [])
            messages = load_res.get('messages', [])
            res['messages'].extend(messages)
            for message in messages:
                if message.get('type') in ['warning', 'error']:
                    if message.get('record'):
                        self.logger.error("record : %s" % (message['record']))
                    if message.get('message'):
                        self.logger.error("message : %s" % (message['message']))
                else:
                    self.logger.info(message)
            stop_batch = datetime.now()
            self.logger.info("\t\t\tBatch time %s ( %sms per object)" % (
                stop_batch - start_batch, ((stop_batch - start_batch) / len(load_data)).microseconds / 1000))
        stop = datetime.now()
        self.logger.info("\t\t\tTotal time %s" % (stop - start))
        return res

    def create(self, model, values, context=None):
        if isinstance(values, dict):
            values = [values]
        res = self.execute_odoo(model, 'create', [values],  {'context': context or self._context})
        if isinstance(res, int):
            res = [res]
        res_values = [{'id': r} | values[i] for i, r in enumerate(res)]
        return self.values_list_to_records(model, res_values)

    def unlink(self, model, values, context=None):
        return self.execute_odoo(model, 'unlink', [values],  {'context': context or self._context})

    def unlink_domain(self, model, domain, context=None):
        values = self.search_ids(model, domain)
        return self.unlink(model, values, context)

    def create_attachment(self, name, datas, res_model, res_id=False, context=None):
        values = {
                    'name': name,
                    'datas': datas,
                    'res_model': res_model}
        if res_id:
            values['res_id'] = res_id
        return self.create('ir.attachment', values,  context)

    def create_attachment_from_local_file(self, file_path, res_model, res_id=False,
                                          name=False, encode=False, context=None):
        datas = self.get_local_file(file_path, encode)
        file_name = name or os.path.basename(file_path)
        return self.create_attachment(file_name, datas, res_model, res_id, context)

    def get_ir_model_data(self, model):
        return self.search('ir.model.data', [('model', '=', model)],
                    fields=['module', 'name', 'model', 'res_id'])

    def get_id_ref_dict(self, model):
        """
        Returns a dict with id as key and xmlid as value
        :param model: Model name
        :return: {894: 'base.module_account', ...}
        """
        model_datas = self.get_ir_model_data(model)
        return dict([(data.res_id, '%s.%s' % (data.module, data.name)) for data in model_datas])

    def get_xmlid_dict(self, model):
        """
        Returns a dict with xmlid as key and id as value
        :param model: Model name
        :return: {'base.module_account': 894, ...}
        """
        model_datas = self.get_ir_model_data(model)
        return dict([('%s.%s' % (data.module, data.name), data.res_id) for data in model_datas])

    def get_id_ref_list(self, model):
        """
        Returns a dict with record id as key and a xmlid list as value
        :param model: Model name
        :return: {'894': ['base.module_account', ...], ...}
        """
        xmlid_dict = self.get_xmlid_dict(model)
        return dict([(id, [xmlid for xmlid, res_id in xmlid_dict.items() if res_id == id]) for id in xmlid_dict.values()])

    def get_fields(self, model_name, fields=None, attributes=None):
        model = self.model(model_name)
        if model._fields and not fields and not attributes:
            return model._fields
        if model._fields and fields:
            return {f:model._fields[f] for f in fields if f in model._fields}
        params = [fields or []]
        if attributes:
            params.append(attributes)

        res_fields = self.execute_odoo(model_name, 'fields_get', params)
        if not fields and model and not model._fields_loaded:
            model._fields = res_fields
        return res_fields

    def _get_xmlid_cache(self, model_name, res_id):
        if not 'ir.model.data' in self._models:
            return
        cache = self._get_model_cache('ir.model.data')
        ir_model_datas = list(filter(lambda x: x['model'] == model_name and x['res_id'] == res_id, cache))
        if ir_model_datas:
            return '{0}.{1}'.format(ir_model_datas[0]['module'], ir_model_datas[0]['name'])

    def _get_object_reference_cache(self, xml_id):
        if not 'ir.model.data' in self._models:
            return
        module, name = xml_id.split('.')
        cache = self._get_model_cache('ir.model.data')
        ir_model_datas = list(filter(lambda x: x['module'] == module and x['name'] == name, cache))
        if ir_model_datas:
            return [ir_model_datas[0]['model'], ir_model_datas[0]['res_id']]

    def _get_model_cache(self, model_name):
        cache = self.model(model_name)._cache
        return [cache[res_id] for res_id in cache.keys()]

    def clear_cache(self):
        for model in self._models:
            self.model(model)._cache = {}

    def print_query_count(self):
        self.logger.info('Query count : %s', self.query_count)
        for method in self.method_count:
            self.logger.info('Method Count %s : %s', method, self.method_count[method])

    def get_method_count(self, method, model=None):
        if model:
            method_values = self.method_count.get(method, {})
            if method_values:
                return method_values.get(model, 0)
            else:
                return sum(method_values.values())
        return 0

    def import_csv(self, file_path, model, **kwargs):
        return self._file_import.import_csv(file_path, model, **kwargs)

    def import_xls(self, file_path, model, **kwargs):
        return self._file_import.import_xls(file_path, model, **kwargs)

    def import_xlsx(self, file_path, model, **kwargs):
        return self._file_import.import_xlsx(file_path, model, **kwargs)


# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import json
from .model import OdooModel


class OdooJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder for OdooRecord and OdooRecordSet"""

    def default(self, obj):
        if hasattr(obj, 'to_dict'):
            return obj.to_dict()
        if isinstance(obj, (list, tuple)) and hasattr(obj, 'ids'):  # Handle OdooRecordSet
            return [record.to_dict() if hasattr(record, 'to_dict') else dict(record) for record in obj]
        return super().default(obj)


# Monkey patch json.dumps to use our custom encoder by default
_original_dumps = json.dumps


def dumps(*args, **kwargs):
    if 'cls' not in kwargs:
        kwargs['cls'] = OdooJSONEncoder
    return _original_dumps(*args, **kwargs)


json.dumps = dumps


class OdooRecord(object):
    _odoo = None
    _model = None
    _field = ''
    _parent_model = None
    _xmlid = ''

    _updated_values = {}
    _initialized_fields = []

    def __init__(self, odoo, model, values: dict, field='', parent_model=None, **kwargs):
        self._values = {}
        self._updated_values = {}
        self._initialized_fields = []
        self._odoo = odoo
        self.id = False
        if model:
            self._model = model
            self._odoo = self._model._odoo
        if field:
            self._field = field
        if parent_model:
            self._parent_model = parent_model

        self.set_values(values, update_cache=False, resolve_xmlids=kwargs.get('resolve_xmlids', True))

    def __str__(self):
        if self._model:
            if hasattr(self, 'id') and self.id:
                return "%s(%s)" % (self._model, self.id)
            if 'name' in self._values:
                return "%s(%s)" % (self._model, self.name)
            elif self._xmlid:
                return "%s(%s)" % (self._model, self._xmlid)
        elif self._field:
            if hasattr(self, 'id'):
                return "%s(%s)" % (self._field, self.id)
            if hasattr(self, 'name'):
                return "%s(%s)" % (self._field, self.name)
        return str(self._values)

    def __repr__(self):
        return str(self)

    def __iter__(self):
        """Support for dict() conversion by yielding key-value pairs"""
        for key in self.get_attributes():
            yield key, self.__dict__[key]

    def get_attributes(self):
        for key in list(self.__dict__.keys()):
            if key.startswith('_'):
                continue
            yield key

    def __getitem__(self, key):
        if isinstance(key, str):
            return getattr(self, key)

    def get(self, key, default=None):
        if key in list(self.__dict__.keys()):
            return getattr(self, key)
        return default

    def to_dict(self):
        """Convert the record to a dictionary with JSON-serializable values"""
        result = {}
        for key in self.get_attributes():
            value = getattr(self, key)
            if isinstance(value, OdooRecord):
                result[key] = value.to_dict()
            elif hasattr(value, 'to_dict'):
                result[key] = value.to_dict()
            elif isinstance(value, dict):
                result[key] = {k: v.to_dict() if hasattr(v, 'to_dict') else v for k, v in value.items()}
            elif isinstance(value, (list, tuple)):
                result[key] = [
                    item.to_dict() if isinstance(item, OdooRecord) else item
                    for item in value
                ]
            else:
                result[key] = value
        return result

    def __bool__(self):
        if hasattr(self, 'id') and self._model:
            return bool(self.id)
        else:
            return any([bool(getattr(self, k)) for k in list(self.__dict__.keys())])

    def __getattribute__(self, name):
        if name.startswith('_'):
            return super().__getattribute__(name)
        if not self._model and name in self._values:
            return getattr(self, name)
        if name not in self._values and self._model:
            if not self._model._fields_loaded:
                self._model.load_fields_description()
            if name in self._model._fields:
                res = self.read([name])
            else:
                res = super().__getattribute__(name)
        else:
            res = super().__getattribute__(name)
        if isinstance(res, dict):
            return OdooRecord(self._odoo, None, res)
        if res and isinstance(res, list) and not isinstance(res[0], OdooRecord):
            return self._odoo.values_list_to_records(None, res)
        return res

    def __setattr__(self, name, value):
        if name.startswith('_') or name == 'id':
            return super().__setattr__(name, value)
        if name not in self._values and self._model:
            if not self._model._fields_loaded:
                self._model.load_fields_description()
        if name in self._values and name in self._initialized_fields and value != self._values[name]:
            self._updated_values[name] = value
            return super().__setattr__(name, value)
        if name in self._values and name not in self._initialized_fields:
            self._updated_values[name] = value
            res = super().__setattr__(name, value)
            self._initialized_fields.append(name)
            return res
        return super().__setattr__(name, value)

    def __setitem__(self, key, value):
        if isinstance(key, str):
            return self.__setattr__(key, value)
        return super().__setitem__(key, value)

    def __hash__(self):
        return hash((self.id, self._model))

    def __eq__(self, other):
        if hasattr(other, 'id') and hasattr(other, '_model'):
            return self.id == other.id and self._model == other._model
        return super().__eq__(other)

    @property
    def _read_fields(self):
        return [k for k in self.__dict__.keys() if not k.startswith('_')]

    def _update_cache(self):
        if self._model:
            self._model._update_cache(self._values['id'], self._values)

    def set_values(self, values, update_cache=True, resolve_xmlids=True):
        values = values.copy()
        self._values.update(values)
        self._handle_id_and_xmlid(values)
        if self._model and update_cache:
            self._update_cache()
        self._process_related_values(values, resolve_xmlids)

    def _handle_id_and_xmlid(self, values):
        if not values.get('id', False):
            self._updated_values = values
        if '/id' in values:
            self._xmlid = values.pop('/id', None)
        if 'id' in values and isinstance(values['id'], str):
            self._xmlid = values.pop('id', None)

    def _process_related_values(self, values, resolve_xmlids=True):
        value_type = 'id'
        for key, value in values.items():
            if '.' in key:
                field_name = key.split('.')[0]
            elif '/' in key:
                value_type = 'xmlid'
                field_name = key.split('/')[0]
            else:
                field_name = key

            if self._model:
                field = self._model.get_field(field_name)
            else:
                field = None

            if field and isinstance(value, list) and len(value) == 2 and isinstance(value[0], int) and isinstance(
                    value[1], str):
                self._handle_relation_list(field_name, value, field)
            elif not field:
                if isinstance(value, list):
                    if len(value) == 2 and isinstance(value[0], int) and isinstance(value[1], str):
                        value = OdooRecord(self._odoo, None, {'id': value[0], 'name': value[1]})
                elif isinstance(value, dict):
                    value = OdooRecord(self._odoo, None, value)
                super().__setattr__(key, value)
            elif field.get('type') == 'many2many' and key.endswith('id') or key.endswith('ids'):
                self._values.pop(key, None)
                self._handle_relation_many2many_ids(field_name, value, value_type, resolve_xmlids)
            elif key.endswith('/id') and isinstance(value, str):
                if resolve_xmlids:
                    self._values.pop(key, None)
                    self._handle_relation_xmlid(field_name, value, resolve_xmlids)
            elif key.endswith('.id') and isinstance(value, int):
                self._handle_relation_id(field_name, value)
            elif isinstance(value, dict):
                value = OdooRecord(self._odoo, None, value)
                super().__setattr__(key, value)
            else:
                try:
                    super().__setattr__(key, value)
                except Exception as e:
                    self.logger.error(e)

    def _handle_relation_list(self, key, value, field=False):
        field_name = key
        if not field or not field.get('relation'):
            return
        relation = field.get('relation')
        model = OdooModel(self._odoo, relation)
        if field.get('type') == 'many2one':
            record = OdooRecord(self._odoo, model, {'id': value[0], 'name': value[1]}, field_name, self._model)
        else:
            record = self._odoo.values_list_to_records(relation, [{'id': val} for val in value])
        super().__setattr__(key, record)

    def _handle_relation_many2many_ids(self, field_name, value, value_type, resolve_xmlids=True):
        if isinstance(value, str):
            value = value.split(',')
        if value_type == 'id':
            value = [int(val) for val in value]
        if value_type == 'xmlid' and resolve_xmlids:
            value = [self._odoo.get_ref(val) for val in value]
        field = self._model.get_field(field_name)
        if not field or not field.get('relation'):
            return
        relation = field.get('relation')
        record = self._odoo.values_list_to_records(relation, [{'id': val} for val in value])
        super().__setattr__(field_name, record)

    def _handle_relation_xmlid(self, field_name, value, resolve_xmlids):
        field = self._model.get_field(field_name)
        if not field.get('relation'):
            return
        if resolve_xmlids:
            res_id = self._odoo.get_ref(value)
        else:
            res_id = value

        model = OdooModel(self._odoo, field.get('relation'))
        record = OdooRecord(self._odoo, model, {'id': res_id}, field_name, self._model)
        record._xmlid = value
        super().__setattr__(field_name, record)
        if resolve_xmlids:
            self._values[f'{field_name}.id'] = record.id
        else:
            self._values[f'{field_name}/id'] = record._xmlid

    def _handle_relation_id(self, field_name, value):
        field = self._model.get_field(field_name)
        if not field.get('relation'):
            return
        model = OdooModel(self._odoo, field.get('relation'))
        record = OdooRecord(self._odoo, model, {'id': value}, field_name, self._model)
        super().__setattr__(field_name, record)
        self._values[f'{field_name}.id'] = record.id

    def read(self, fields=None, no_cache=False):
        if not self._model._fields_loaded:
            self._model.load_fields_description()
        if self.id in self._model._cache and not no_cache:
            res = self._model._cache[self.id]
            # check if all fields are in res dict
            if any(field not in res for field in fields):
                res.update(self._read(fields))

            self.set_values(res)
        else:
            if not fields:
                fields = self._model.get_fields_list()
            res = self._read(fields)
            if res:
                self.set_values(res)

    def _read(self, fields):
        res = self._model._read(self.id, fields)
        if res:
            return res[0]

    def save(self):
        values = self.get_update_values()
        if self._xmlid:
            res = self._model.load(list(values.keys()), [list(values.values())])
            if res.get('ids'):
                self.id = res.get('ids')[0]
            return
        if self.id:
            self._model.write(self.id, self._updated_values)
            self._updated_values = {}
        else:
            self.id = self._odoo.create(self._model.model_name, self._values)[0].id
            self._initialized_fields = list(self._values.keys())

    def write(self, values):
        self._model.write(self.id, values)
        self._values.update(values)
        self.__dict__.update(values)

    def refresh(self):
        self.read(self._initialized_fields, no_cache=True)

    def get_update_values(self):
        if '/id' in self._values:
            self._xmlid = self._values['/id']
            self._values.pop('/id', None)

        value_id = self._values.get('id', False)
        if isinstance(value_id, str):
            self._xmlid = value_id
            self.id = None
            self._values['id'] = None

        if self.id:
            values = self._updated_values
            values['.id'] = self.id
        else:
            values = self._values
            if self._xmlid:
                values['id'] = self._xmlid
            else:
                values['id'] = None

        return values

    def unlink(self):
        if self.id:
            self._model.unlink(self.id)
            self._model._cache.pop(self.id, None)
        self.id = None



from ._fixtures import *

from s6r_odoo.odoo import OdooConnection
from s6r_odoo.model import OdooModel
from s6r_odoo.record import OdooRecord
from s6r_odoo.record_set import OdooRecordSet


def test_values_to_record(odoo):
    model_name = 'res.partner'

    partner_id = odoo.values_to_record(model_name,123)
    assert isinstance(partner_id, OdooRecord)
    assert partner_id.id == 123

    partner_values = {
        'name': 'Jean-Test',
        'email': 'jean@test.cool'
    }
    partner_id = odoo.values_to_record(model_name,partner_values)
    assert isinstance(partner_id, OdooRecord)
    assert partner_id.name == 'Jean-Test'

    #TODO test cache

def test_get_xmlid_dict(odoo):
    xmlids = odoo.model('ir.module.module').get_xmlid_dict()

    assert 'base.module_account' in xmlids

def test_get_id_ref_dict(odoo):
    id_refs = odoo.model('ir.module.module').get_id_ref_dict()

    #check if dict has at least a key '1' containing a string with a '.'
    assert len(id_refs.get(1).split('.')) == 2

def test_get_ir_model_data(odoo_multi_versions):
    odoo = odoo_multi_versions
    ir_model_data = odoo.model('ir.module.module').get_ir_model_data()

    assert isinstance(ir_model_data, OdooRecordSet)
    assert ir_model_data._model.model_name == 'ir.model.data'




from ._fixtures import *

from s6r_odoo.odoo import OdooConnection
from s6r_odoo.model import OdooModel
from s6r_odoo.record import OdooRecord
from s6r_odoo.record_set import OdooRecordSet

def test_save(odoo):

    partner_values_list = [
        {
            '/id':'external_import.res_partner_sauvequipeut',
            'name': 'Sauve qui peut',
            'email': 'sqp@test.com',
            'parent_id/id': 'external_import.res_partner_lamaisonmere',

        },
        {
            '/id': 'external_import.res_partner_sauvezlesdauphins',
            'name': 'Sauvez les dauphins',
            'email': 'sld@test.com',
            'parent_id/id': 'external_import.res_partner_lamaisonmere',
        },
        {
            '/id': 'external_import.res_partner_lamaisonmere',
            'name': 'Maison Mère',
            'email': 'mm@test.com',
            'parent_id/id': '',
        },
    ]
    partner_ids = odoo.values_list_to_records('res.partner', partner_values_list)

    #save without hierarchy first
    partner_ids.save(ignore_fields=['parent_id/id'])

    assert len(partner_ids[0]._initialized_fields) == 2
    assert all([f in ['name', 'email'] for f in partner_ids[0]._initialized_fields])
    assert list(partner_ids[0]._updated_values.keys()) == ['parent_id/id']

    #save with hierarchy
    partner_ids.save()

    assert len(partner_ids[0]._initialized_fields) == 3
    assert all([f in ['name', 'email', 'parent_id'] for f in partner_ids[0]._initialized_fields])
    assert partner_ids[0]._updated_values == {}



import logging
import time
import random
import string
from src.s6r_odoo import OdooConnection

def get_random_string(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for _ in range(length))


def test_local():
    logging.basicConfig()
    logger = logging.getLogger("test")

    odoo = OdooConnection(url='http://odoo_project.localhost',
                          dbname='project',
                          user='admin',
                          password='admin', debug_xmlrpc=False, logger=logger)

    xmlid_dict = odoo.model('ir.module.module').get_xmlid_dict()
    logger.info('ir.module.module XMLIDs : %s', xmlid_dict)
    id_ref_dict = odoo.model('ir.module.module').get_id_ref_dict()
    logger.info('ir.module.module ID refs : %s', id_ref_dict)
    ir_model_data = odoo.model('ir.module.module').get_ir_model_data()
    logger.info('ir.module.module ir_model_data : %s', ir_model_data)

    object_reference = odoo.get_object_reference('base.module_account')
    logger.info("base.module_account get_object_reference --> ['ir.module.module', id] : %s", object_reference)
    res_id = odoo.get_ref('base.module_account')
    logger.info("base.module_account get_ref --> id : %s", res_id)
    module_account = odoo.ref('base.module_account')
    msg = 'base.module_account ref --> OdooRecord :\nModule: %s (%s)\nDescription: %s'
    logger.info(msg, module_account.display_name,
                module_account.name,
                module_account.description)

    odoo.get_xml_id_from_id('ir.module.module', module_account.id)

    odoo.print_query_count()

    partner_fields = odoo.execute_odoo('res.partner', 'fields_get', [['name'], None])
    partner_fields2 = odoo.model('res.partner').execute('fields_get', ['name'])

    country_fields = odoo.model('res.country').get_fields([])
    country_required_fields = ', '.join(
        country_fields[key]['name'] for key in country_fields if country_fields[key]['required'])
    logger.info('Country required fields : %s', country_required_fields)

    # Script optimised to access license and category_id field
    start_time = time.time()
    module_ids = odoo.model('ir.module.module').search([('author', 'ilike', 'Odoo')],
                                                       fields=['display_name', 'category_id', 'license'],
                                                       order='create_date desc, name', limit=100)
    optimized_query_time = time.time() - start_time
    start_time = time.time()
    logger.info('Module categories : %s', ', '.join(set([module.category_id.name for module in module_ids])))
    for module_id in module_ids:
        module_license = module_id.license if module_id['license'] else 'Unknown'
        logger.info('* Module %s : %s', module_id.display_name, module_id.license)
        continue
    optimized_time = time.time() - start_time
    logger.info("Optimised Script : %s seconds (first query : %s seconds)", optimized_time, optimized_query_time)

    # Script NOT optimised to access to licence and category_id field
    # start_time = time.time()
    # module_ids = odoo.model('ir.module.module').search([('author', 'ilike', 'Odoo')],
    #                                                    fields=['display_name'],
    #                                                    order='create_date desc, name', limit=10)
    # not_optimized_query_time = time.time() - start_time
    # start_time = time.time()
    # logger.info('Module categories : %s', ', '.join(set([module.category_id.name for module in module_ids])))
    # for module_id in filter(lambda x: x.category_id.name != 'Scalizer', module_ids): # module_ids:
    #     module_license = module_id.license if module_id['license'] else 'Unknown'
    #     logger.info('* Module %s : %s', module_id.display_name, module_id.license)
    #     continue
    # not_optimized_time = time.time() - start_time
    # logger.info("NOT Optimised Script : %s seconds (first query : %s seconds)", not_optimized_time, not_optimized_query_time)


    # XMLIDs Tests
    # logger.info('Module XMLIDs : %s', odoo.get_xmlid_dict('ir.module.module'))

    object_id = odoo.get_ref('base.module_account')
    logger.info('base.module_account get_ref --> id : %s', object_id)
    module_account_lazy = odoo.model('ir.module.module').read(object_id, ['name', 'description'])

    logger.info('Module (LAZY): %s\nCategory: %s\nDescription: %s\n',
                module_account_lazy['name'],
                module_account_lazy['category_id'].name,
                module_account_lazy['description'])

    # test cache
    origin_email = 'test@example.com'
    email1 = 'john@example.com'
    partner_id = odoo.ref('base.main_partner')
    partner_id.email = email1
    partner_id.save()
    if partner_id.email != email1:
        raise Exception('Email %s not updated' % email1)
    email2 = 'sophie@example.com'
    partner_id.write({'email': email2})
    if partner_id.email != email2:
        raise Exception('Email %s not updated' % email2)

    partner_id2 = odoo.model('res.partner').read(partner_id.id)
    if partner_id2.email_formatted and email2 in partner_id2.email_formatted:
        partner_id.write({'email': origin_email})
        raise Exception('Cache not used with read')

    partner_id2 = odoo.ref('base.main_partner')
    if partner_id2.email_formatted and email2 in partner_id2.email_formatted:
        partner_id.write({'email': origin_email})
        raise Exception('Cache not used with ref')

    partner_id3 = odoo.ref('base.main_partner', no_cache=True)
    if email2 not in partner_id3.email_formatted:
        partner_id.write({'email': origin_email})
        raise Exception('Cache used with ref')
    partner_id3.write({'email': email1})
    partner_id3.refresh()
    if email1 not in partner_id3.email_formatted:
        partner_id.write({'email': origin_email})
        raise Exception('Refresh failed')

    partner_id.write({'email': origin_email})

    partner_ids = odoo.model('res.partner').search([], fields=['name', 'website'])
    for partner_id in partner_ids:
        partner_id.website = 'https://www.%s.com' % get_random_string(8)
    partner_ids.save()

    partner_ids = odoo.model('res.partner').search([('name', '=', 'TO_REMOVE')], fields=['name'])
    if partner_ids:
        partner_ids.unlink()

    new_partner_values_list = [{'name': "TO_REMOVE", 'website': get_random_string(8)} for i in range(5000)]
    new_partner_ids = odoo.model('res.partner').load_batch(new_partner_values_list)
    new_partner_ids.unlink()

    odoo.print_query_count()

    # Test record creation from dict
    partner_tag1_values = {
        'name': 'Test tag without xmlid',
        'color': 2,
    }
    partner_tag1_id = odoo.model('res.partner.category').values_to_record(partner_tag1_values, update_cache=False)
    # this record will be created using "create" method
    partner_tag1_id.save()

    partner_tag2_values = {
        '/id': f'external_config.res_partner_category_test',
        'name': 'Test tag with xmlid',
        'color': 10,
    }
    partner_tag2_id = odoo.model('res.partner.category').values_to_record(partner_tag2_values, update_cache=False)
    # this record will be created using "load" method
    partner_tag2_id.save()
    # cleanup
    partner_tag1_id.unlink()
    partner_tag2_id.unlink()

    partner_tag_values_list = [{'name': f"TAG {i}", 'color': i} for i in range(10)]
    partner_tag_ids = odoo.model('res.partner.category').values_list_to_records(partner_tag_values_list)
    # load the records without xmlid
    for partner_tag in partner_tag_ids:
        partner_tag.computed_value = partner_tag.color % 2 == 0

    filtered_tag_ids = partner_tag_ids.filtered(lambda x: x.computed_value)
    filtered_tag_ids2 = partner_tag_ids.filtered(computed_value=True, name='TAG 8')

    partner_tag_ids.save()
    partner_tag_ids.unlink()

    partner_tags_list2 = [{'/id': f"external_config.res_partner_category_{i}", 'name': f"TAG {i}", 'color': i - 10} for
                          i in range(10, 20)]
    partner_tag_ids2 = odoo.model('res.partner.category').values_list_to_records(partner_tags_list2)
    # load the records with xmlid
    partner_tag_ids2.save()
    partner_tag_ids2.unlink()

    data_ids = odoo.search_ir_model_data([('module', '=', 'base')])
    models = data_ids.mapped('model')
    logger.info('Models : %s', ', '.join(models))

    partner_ids = odoo.model('res.partner').search([], fields=['country_id', 'name'])
    country_codes = partner_ids.mapped('country_id.code')
    logger.info('Country codes : %s', ', '.join(country_codes))

    data_ids = odoo.search_ir_model_data([('module', '=', 'base')])
    if data_ids:
        models = data_ids.mapped('model')
        for model in models:
            model_data_ids = data_ids.filtered(model=model)
            record_ids = odoo.model(model).read(model_data_ids.mapped('res_id'))
            # logger.info(record_ids)
            break

    data_users = [{'id': f'external_config.user_test_01',
                   'name': 'Lastname Firstname TEST1', 'login': 'test1@test.fr',
                   'title.id': 1, 'company_ids.id': [1, 3], 'active': True},
                  {'id': f'external_config.user_test_02',
                   'name': 'Lastname Firstname TEST2', 'login': 'test2@test.fr',
                   'title.id': 1, 'company_ids.id': [1, 3], 'active': True}]
    odoo.model('res.users').load_batch(data_users)

if __name__ == '__main__':
    test_local()


try:
    from ._fixtures import *
except ImportError:
    pass


def _test_load_data(odoo, data):
    res = odoo.model('res.partner').load_batch(data)
    if res and res.get('messages'):
        for message in res.get('messages'):
            odoo.logger.error(message)
        return


def test_load_xmlids(odoo):
    category_values = [{'id': 'external_config.category_test_01', 'name': 'Category 01'},
                       {'id': 'external_config.category_test_02', 'name': 'Category 02'}]
    res = odoo.model('res.partner.category').load_batch(category_values)
    if res and res.get('messages'):
        for message in res.get('messages'):
            odoo.logger.error(message)
        return
    assert res and res.get('ids')
    assert len(res.get('ids')) == 2
    assert odoo.query_count == 1
    odoo.reset_count()
    category_ids = res.get('ids')
    category_ids_str = ','.join([str(i) for i in category_ids])
    category_xmlids_str = ','.join([v['id'] for v in category_values])

    title_ids = odoo.model('res.partner.title').read(['base.res_partner_title_madam', 'base.res_partner_title_mister'],
                                                     ['name'])
    odoo.reset_count()
    data_partners_with_xmlids = [{'id': f'external_config.partner_test_01',
                      'name': 'Lastname Firstname 01',
                      'title/id': 'base.res_partner_title_madam',
                      'category_id/id': category_xmlids_str},
                     {'id': f'external_config.partner_test_02',
                      'name': 'Lastname Firstname 02',
                      'title/id': 'base.res_partner_title_mister',
                      'category_id/id': category_xmlids_str}]

    _test_load_data(odoo, data_partners_with_xmlids)
    partner_ids = odoo.model('res.partner').values_list_to_records(data_partners_with_xmlids)
    assert partner_ids[0].category_id[0].id > 0

    # 2 queries to get titles xmlids (categories xmlids are in cache)
    assert odoo.get_method_count('search_read', 'ir.model.data') == 2

    odoo.reset_count()
    partner_ids = odoo.model('res.partner').values_list_to_records(data_partners_with_xmlids)

    # Use the xmlid cache to get the id
    assert partner_ids[0].title.id == 1
    assert odoo.get_method_count('search_read', 'ir.model.data') == 0

    # Clear the cache and get the title xmlids again
    odoo.clear_cache()
    odoo.reset_count()
    partner_ids = odoo.model('res.partner').values_list_to_records(data_partners_with_xmlids)
    assert partner_ids[0].title.id == 1
    assert odoo.get_method_count('search_read', 'ir.model.data') == 4
    partner_ids.save()

    # test without xmlids in values
    odoo.clear_cache()
    odoo.reset_count()
    data_partners = [{'id': f'external_config.partner_test_01',
                      'name': 'Lastname Firstname 01',
                      'title.id': title_ids[0].id,
                      'category_id.id': category_ids_str},
                     {'id': f'external_config.partner_test_02',
                      'name': 'Lastname Firstname 02',
                      'title.id': title_ids[1].id,
                      'category_id.id': category_ids_str}]
    _test_load_data(odoo, data_partners)
    partner_ids = odoo.model('res.partner').values_list_to_records(data_partners)
    partner_ids.save()
    assert partner_ids[0].id > 0
    assert odoo.get_method_count('search_read', 'ir.model.data') == 0

    # Test resolve_xmlids param
    odoo.clear_cache()
    odoo.reset_count()
    partner_ids = odoo.model('res.partner').values_list_to_records(data_partners, resolve_xmlids=False)
    partner_ids.save()
    assert partner_ids[0].id > 0
    assert odoo.get_method_count('search_read', 'ir.model.data') == 0
